/*--------------------------------------------------------------------------------------*/
/* Flash Cards -> TI Substitute (FCTS)                                                  */
/* Programmer: Austin Chu (qarv@yahoo.com)                                              */
/* Programmed in C for the TI-GCC compiler                                              */
/* Version: 0.0.3 (alpha-3, archive support)                                            */
/*                                                                                      */
/* FCTS.c - For now, the whole program                                                  */
/*--------------------------------------------------------------------------------------*/

#define RETURN_VALUE                        // Return Pushed Expression

#define OPTIMIZE_ROM_CALLS                  // Use ROM Call Optimization

#define SAVE_SCREEN                         // Save/Restore LCD Contents

#include <tigcclib.h>                       // Include All Header Files

short _ti89;                                // Produce .89Z File

#define THIS_VERSION 0x0003                 // The version number of this compilation

// Function prototypes
unsigned char* FetchFile(HANDLE* fileHandle);
short RunThruCards(unsigned char* fileTop);
void DrawCard(unsigned char* curPos, short curIndex, short numCards);
unsigned short GetKeypress(void);

// Main Function
void _main(void)
{
        // The handle of the file (H_NULL if originally locked)
    HANDLE fileHandle;
    unsigned char* fileTop;                 // Pointer to the top of the file
    unsigned short fileVersion;             // The file format version
    short lastCard;                         // The last card displayed
    
    fileTop = FetchFile(&fileHandle);       // Get the file
    
    // Clean up the expressions stack
    while(GetArgType(top_estack) != END_TAG)
        top_estack = next_expression_index(top_estack);
    top_estack--;
    
    if(!fileTop)                            // There was an error
    {
        push_longint(-1);
        return;
    }
    
        // Get the format version of the file
    fileVersion = *((unsigned short*)fileTop + 1);
    if(fileVersion > THIS_VERSION)
    {                                       // The file version is not supported
        DlgMessage("FCTS Error",
                   "This version of FCTS is outdated.  Please get a new version.",
                   BT_OK, BT_NONE);
        push_longint(-1);
        return;
    }
    
    lastCard = RunThruCards(fileTop);
    
    // Unlock file if originally not locked
    if(fileHandle)
        HeapUnlock(fileHandle);
    
    push_longint(lastCard);
    return;
}

/* Fetches and locks the flash cards file, error checking along the way
   Arguments: a pointer to a handle which stores the fcts file's handle
                  (NOTE: will be H_NULL if the file was originally locked)
   Returns: pointer to top of the file if successful, NULL if there is an error */
unsigned char* FetchFile(HANDLE* fileHandle)
{
    ESI argPtr = top_estack;                // Used in getting the argument
    HSym theFile;                           // The HSym of the file
    unsigned char* fileTop;                 // Pointer to the top of the file
    unsigned char* fileBtm;                 // Pointer to the bottom of the file
    
    if(GetArgType(argPtr) != STR_TAG)
    {                                       // The argument is not a string
        DlgMessage("FCTS Error", "The argument must be a file name enclosed in quotes.",
                   BT_OK, BT_NONE);
        return NULL;
    }
    
    // Get the HSym of the file indicated by the argument
    theFile = SymFind(GetSymstrArg(argPtr));
    if(theFile.folder == 0)
    {                                       // Error in the SymFind
        DlgMessage("FCTS Error", "The argument file not found.", BT_OK, BT_NONE);
        return NULL;
    }
    
    // Lock, if needed, and get a pointer to the file
    *fileHandle = DerefSym(theFile)->handle;
    if(HeapGetLock(*fileHandle))
    {                                       // The file was locked
        fileTop = HeapDeref(*fileHandle);
        *fileHandle = H_NULL;
    }
    else                                    // The file wasn't locked
        fileTop = HLock(*fileHandle);
    
        // Get bottom of the file
    fileBtm = fileTop + *(unsigned short*)fileTop + 1;
    if(*fileBtm != OTH_TAG || strcmp(fileBtm - 5, "fcts"))
    {                                       // File is not a fcts file
        if(*fileHandle)                     // File was not locked originally
            HeapUnlock(*fileHandle);
        DlgMessage("FCTS Error", "The argument file is not the right type.",
                   BT_OK, BT_NONE);
        return NULL;
    }
    
    // File passes all tests
    return fileTop;
}

/* The meat of the program; controls most interaction with the user
   Arguments: a pointer to the top of the fcts file
   Returns: the index of the last viewed card */
short RunThruCards(unsigned char* fileTop)
{
        // Get the total number of cards
    short numCards = *((short*)fileTop + 2);
        // The index of the current card (if pos->front, neg->back);
    short curIndex = 1;
    unsigned char* curPos = fileTop + 7;    // The current position in the file
    unsigned char newIndex[6] = "";         // Temp string for the new index in KEY_APPS
    HANDLE theDialog;                       // A dialog box for KEY_APPS
    short tmpIndex;                         // Temp number for the new index in KEY_APPS
    short loopIndex;                        // A loop index for KEY_APPS
    
    while(1)
    {
        DrawCard(curPos, curIndex, numCards);
        
        switch(GetKeypress())
        {
            case 344:                       // KEY_RIGHT: continue to next part
                while(*curPos++) { }        // Loop until next string
                if(!(*curPos))
                {                           // End of file hit, reset to first card
                    curIndex = 1;
                    curPos = fileTop + 7;
                }
                else
                    if(curIndex > 0)        // Front of a card, need to flip
                        curIndex = -curIndex;
                    else                    // Back of a card, need to proceed
                        curIndex = -curIndex + 1;
                break;
            
            case 338:                       // KEY_LEFT: return to previous part
                if(curIndex == 1)
                {                           // Loop back to back of last card
                    curIndex = -numCards;
                        // Get the last character of the last card
                    curPos = fileTop + *(unsigned short*)fileTop - 7;
                    while(*curPos--) { }    // Loop until the string before is hit
                    curPos += 2;            // Push forward to the wanted string
                }
                else
                {
                    curPos -= 2;            // Get into string before
                    while(*curPos--) { }    // Loop until 2 strings before is hit
                    curPos += 2;            // Push forward to the wanted string
                    if(curIndex > 0)        // Front of a card, need to go to back of prev
                        curIndex = -(curIndex - 1);
                    else                    // Back of a card, need to go to front
                        curIndex = -curIndex;
                }
                break;
            
            case 265:                       // KEY_APPS: goto a certain card
                // Create a dialog box
                theDialog = DialogNewSimple(126, 37);
                DialogAddTitle(theDialog, "Goto a Card", BT_OK, BT_CANCEL);
                DialogAddRequest(theDialog, 6, 15, "Number of the card:", 0, 5, 6);
                
                // Display the dialog box
                if(DialogDo(theDialog, CENTER, CENTER, newIndex, NULL) == KEY_ENTER)
                {                           // User wants to goto a card
                        // A temporary place for the new index
                    tmpIndex = atoi(newIndex);
                    
                    // Test the new index
                    if(tmpIndex > 0 && tmpIndex <= numCards)
                    {
                        curIndex = tmpIndex;
                        
                        // Set up curPos
                        if(curIndex > numCards / 2)
                        {                   // More than half-way through
                                // Start seeking from the back of the pile
                            curPos = fileTop + *(unsigned short*)fileTop - 7;
                            for(loopIndex = numCards; loopIndex >= curIndex; loopIndex--)
                            {
                                    // Get to the end of the front side
                                while(*curPos--) { }
                                    // Get to the end of the
                                    // back side of the previous card
                                while(*curPos--) { }
                            }
                            curPos += 2;    // Push forward to the wanted string
                        }
                        else
                        {
                                // Start seeking from the top of the pile
                            curPos = fileTop + 7;
                            for(loopIndex = 1; loopIndex < curIndex; loopIndex++)
                            {
                                    // Get to the beginning of the back side
                                while(*curPos++) { }
                                    // Get to the beginning of the
                                    // front side of the next card
                                while(*curPos++) { }
                            }
                        }
                    }
                }
                
                HeapFree(theDialog);
                break;
            
            case KEY_F1:                    // KEY_F1: help
                FillLines2(&(WIN_RECT){0, 92, 159, 92}, &(WIN_RECT){0, 0, 159, 0},
                           &(SCR_RECT){{0, 0, 159, 99}}, A_REVERSE);
                FontSetSys(F_6x8);
                DrawStr(28, 1, "Flash Cards -> TI", A_NORMAL);
                DrawStr(1, 10, "Substitute (FCTS) v. 0.0.3", A_NORMAL);
                DrawStr(7, 25, "Programmed by Austin Chu", A_NORMAL);
                DrawStr(31, 34, "(qarv@yahoo.com)", A_NORMAL);
                DrawStr(0, 57, "[RIGHT] -- go forwards", A_NORMAL);
                DrawStr(0, 66, "[LEFT]  -- go backwards", A_NORMAL);
                DrawStr(0, 75, "[APPS]  -- seek to card #", A_NORMAL);
                DrawStr(0, 84, "[ESC]   -- exit FCTS", A_NORMAL);
                GetKeypress();
                break;
            
            case KEY_ESC:                   // KEY_ESC: exit the program
                return abs(curIndex);
        }
    }
}

/* Draws to the screen the card face
   Arguments: a pointer to the string to be displayed
              the index of the current card
              the total number of cards
   Returns: nothing */
void DrawCard(unsigned char* curPos, short curIndex, short numCards)
{
    unsigned char buffer[27];               // A buffer for various strings
    short bufIndex = -1;                    // An index for the buffer
    short lastSpace = 0;                    // A marker for the last space
    short topLine = 1;                      // For keeping track of top of next line

    // Set up the card index string
    if(curIndex > 0)
        sprintf(buffer, "Front, %i of %i", curIndex, numCards);
    else
        sprintf(buffer, "Back, %i of %i", abs(curIndex), numCards);

    ClearScreen();
    
    // Draw the bottom bar
    FontSetSys(F_4x6);
    DrawStr(0, 94, buffer, A_NORMAL);
    DrawStr(115, 94, "[F1] for keys", A_NORMAL);
    ScrRectFill(&(SCR_RECT){{0, 93, 159, 99}}, &(SCR_RECT){{0, 0, 159, 99}}, A_XOR);
    
    // Draw the card contents
    FontSetSys(F_6x8);
    
    // Keep feeding stuff from the file to buffer until NULL (end of entry) is found
    while((buffer[++bufIndex] = *curPos++))
    {
        if(buffer[bufIndex] == ' ')         // Update the last space
            lastSpace = bufIndex;
        
        if(bufIndex == 26)
        {                                   // Enough characters for a line read
            if(lastSpace == 0)
            {                               // There were no spaces, so display all
                buffer[26] = '\0';
                DrawStr(0, topLine, buffer, A_NORMAL);
                curPos--;                   // Reset the curPos
            }
            else
            {                               // Display characters up to the last space
                buffer[lastSpace] = '\0';
                DrawStr(0, topLine, buffer, A_NORMAL);
                curPos -= 26 - lastSpace;   // Reset the curPos
            }
            
            // Reset/update the counters
            lastSpace = 0;
            bufIndex = -1;
            topLine += 9;
        }
    }
    
    DrawStr(0, topLine, buffer, A_NORMAL);  // Draw the last line
    
    return;
}

/* Function that gets the user's keyboard input
   (Note: This is based largely on GetUserInput() from the TICT-Explorer v1.00.  That
          program was written by the wonderful TI-Chess Team (http://tict.ticalc.org).)
   Arguments: none
   Returns: the key code of the pressed key, minus the repeator flag */
unsigned short GetKeypress(void)
{
    void* kbq = kbd_queue();
    unsigned short key;
    
    OSTimerRestart(APD_TIMER);              // Restart the APD timer
    
    while(1)
    {
        while(OSdequeue(&key, kbq))
        {                                   // Wait for a key/key interrupt
            idle();
            if(OSTimerExpired(APD_TIMER))
            {                               // APD timer ran out, so automatic shutdown
                off();
                OSTimerRestart(APD_TIMER);
            }
        }
        
        if(key == 4363 || key == 16651)
        {                                   // User hit [2nd] + [ON] or [diamond] + [ON]
            off();
            OSTimerRestart(APD_TIMER);
        }
        else                                // We've got a key to return
                // Map out the repeat flag
            return key & ~(unsigned short)0x0800;
    }
}
